﻿/*
 * A wrapper to some FFTW methods (http://www.fftw.org/).
 * Based on Tamás Szalay's wrapper classes (http://www.sdss.jhu.edu/~tamas/bytes/fftwcsharp.html).
 * 
 */
 
using System;
using System.Runtime.InteropServices;

namespace TetheredSun.Spectral.Fftw
{
	/// <summary>
	/// Description of Interop.
	/// </summary>
	internal static class Interop
	{
		/// <summary>
		/// Allocates FFTW-optimised unmanaged memory.
		/// </summary>
		/// <param name="length">Amount to allocate, in bytes</param>
		/// <returns>Pointer to allocated memory</returns>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_malloc", ExactSpelling = true)]
		internal static extern IntPtr AllocateMemory(int length);

		/// <summary>
		/// Deallocates memory allocated by FFTW AllocateMemory.
		/// </summary>
		/// <param name="memoryPointer">Pointer to memory to release</param>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_free", ExactSpelling = true)]
		internal static extern void FreeMemory(IntPtr memoryPointer);

		/// <summary>
		/// Deallocates an FFTW plan and all associated resources.
		/// </summary>
		/// <param name="plan">Pointer to the plan to release</param>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_destroy_plan", ExactSpelling = true)]
		internal static extern void DestroyPlan(IntPtr plan);

		/// <summary>
		/// Clears all memory used by FFTW, resets it to initial state. Does not replace DestroyPlan and FreeMemory.
		/// </summary>
		/// <remarks>After calling Cleanup, all existing plans become undefined, and you should not 
		/// attempt to execute them nor to destroy them. You can however create and execute/destroy new plans, 
		/// in which case FFTW starts accumulating wisdom information again. 
		/// Cleanup does not deallocate your plans; you should still call DestroyPlan for this purpose.</remarks>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_cleanup", ExactSpelling = true)]
		internal static extern void Cleanup();
		
		/// <summary>
		/// Executes an FFTW plan, provided that the input and output arrays still exist.
		/// </summary>
		/// <param name="plan">Pointer to the plan to execute</param>
		/// <remarks>execute (and equivalents) is the only function in FFTW guaranteed to be thread-safe.</remarks>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_execute", ExactSpelling = true)]
		internal static extern void Execute(IntPtr plan);
		
		/// <summary>
		/// Creates a plan for a 1-dimensional complex-to-complex DFT.
		/// </summary>
		/// <param name="n">The logical size of the transform</param>
		/// <param name="direction">Specifies the direction of the transform</param>
		/// <param name="input">Pointer to an array of 16-byte complex numbers</param>
		/// <param name="output">Pointer to an array of 16-byte complex numbers</param>
		/// <param name="flags">Flags that specify the behaviour of the planner</param>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_plan_dft_1d", ExactSpelling = true)]
		internal static extern IntPtr OneDimensionalComplexToComplex(int n, IntPtr input, IntPtr output, FftwDirection direction, FftwFlags flags);
		
		/// <summary>
		/// Creates a plan for a 1-dimensional real-to-complex DFT.
		/// </summary>
		/// <param name="n">Number of REAL (input) elements in the transform</param>
		/// <param name="input">Pointer to an array of 8-byte real numbers</param>
		/// <param name="output">Pointer to an array of 16-byte complex numbers</param>
		/// <param name="flags">Flags that specify the behaviour of the planner</param>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_plan_dft_r2c_1d", ExactSpelling = true)]
		internal static extern IntPtr OneDimensionalRealToComplex(int n, IntPtr input, IntPtr output, FftwFlags flags);
		
		/// <summary>
		/// Creates a plan for a 1-dimensional complex-to-real DFT.
		/// </summary>
		/// <param name="n">Number of REAL (output) elements in the transform</param>
		/// <param name="input">Pointer to an array of 16-byte complex numbers</param>
		/// <param name="output">Pointer to an array of 8-byte real numbers</param>
		/// <param name="flags">Flags that specify the behavior of the planner</param>
		[DllImport("libfftw3-3.dll", EntryPoint = "fftw_plan_dft_c2r_1d", ExactSpelling = true)]
		internal static extern IntPtr OneDimensionalComplexToReal(int n, IntPtr input, IntPtr output, FftwFlags flags);
	}
}
